theory DT_Transformation
imports
  Q_Relative_Rewriting
  Relative_Rewriting2
  SubList
  Position
  Complexity
  Multihole_Context
begin



(* utils *)

lemma map_sublist_f_g:
  assumes sublst: "sublist xs ys"
  and f_eq_g: "\<forall> x \<in> set xs. f x = g x"
  shows "sublist (map f xs) (map g ys)"
using assms proof (induct ys, (auto iff: Skip Shift)) qed
  
lemma map_sublist:
  assumes sublst: "sublist xs ys" shows "sublist (map f xs) (map f ys)"
  using assms proof (rule map_sublist_f_g[of xs ys f f], simp)  qed

lemma funas_induct [consumes 1, case_names Var Fun, induct type: "term"]: assumes t: "funas_term t \<subseteq> fs"
  and base: "\<And> x. P (Var x)"
  and induct: "\<And> f ts. (f,length ts) \<in> fs \<Longrightarrow> \<lbrakk>\<And> t. t \<in> set ts \<Longrightarrow> funas_term t \<subseteq> fs\<rbrakk> \<Longrightarrow> \<lbrakk>\<And> t. t \<in> set ts \<Longrightarrow> P t\<rbrakk> \<Longrightarrow> P (Fun f ts)"
  shows "P t"
  using t
proof (induct t)
  case (Var x)
  show ?case by (rule base)
next
  case (Fun f ts)
  show ?case
    by (rule induct[OF _ _ Fun(1)], insert Fun, auto)
qed


lemma subterm_poss_conv:
  assumes "p \<in> poss t" and "p = q <#> r" and "t |_ q = s" 
  shows "t |_ p = s |_ r \<and> r \<in> poss s" 
  using assms by (metis poss_append_poss subt_at_append)

lemma subst_image_subterm:
  assumes "x \<in> vars_term (Fun f ss)"
  shows "(Fun f ss) \<cdot> \<sigma> \<rhd> \<sigma> x"
proof -
  have "Fun f ss \<unrhd> Var x" using supteq_Var[OF assms(1)] . 
  hence "Fun f ss \<rhd> Var x" by cases auto
  then show ?thesis using supt_stable by force qed

lemma funposs_fun_conv:
  assumes "p \<in> funposs t"
  shows "\<exists> f ts. t |_ p = Fun f ts"
proof (cases "t |_ p")
  case (Var x)
  have p_in_t: "p \<in> poss t" using funposs_imp_poss[OF assms].
  then have "p \<in> poss t - funposs t" using Var(1) varposs_iff by auto
  then show ?thesis using assms by blast
next
  case (Fun f ts) then show ?thesis by auto
qed

lemma qrstep_r_p_s_conv:
  fixes s t
  assumes step: "(s, t) \<in> qrstep_r_p_s nfs Q R lr p \<sigma>"
  defines [simp]: "C \<equiv> ctxt_of_pos_term p s"
  shows "p \<in> poss s" and "p \<in> poss t" and "s = C\<langle>fst lr \<cdot> \<sigma>\<rangle>" and "t = C\<langle>snd lr \<cdot> \<sigma>\<rangle>" and "NF_subst nfs lr \<sigma> Q" and "ctxt_of_pos_term p t = C"
proof -
  from step have p_in_s: "p \<in> poss s" unfolding qrstep_r_p_s_def by simp
  have p_in_t: "p \<in> poss t" using qrstep_r_p_s_imp_poss[OF step] by simp
  have C: "C = ctxt_of_pos_term p t" using ctxt_of_pos_term_qrstep_below[OF step] by simp

  from p_in_s show "p \<in> poss s".
  show "p \<in> poss t" using qrstep_r_p_s_imp_poss[OF step] by simp
  from step show "s = C\<langle>fst lr \<cdot> \<sigma>\<rangle>" unfolding qrstep_r_p_s_def using ctxt_supt_id[OF p_in_s] by simp
  from C show "t = C\<langle>snd lr \<cdot> \<sigma>\<rangle>" using step unfolding qrstep_r_p_s_def using ctxt_supt_id[OF p_in_t] by simp
  from step show "NF_subst nfs lr \<sigma> Q" unfolding qrstep_r_p_s_def by simp
  from C show "ctxt_of_pos_term p t = C" by simp
qed

lemma qrstep_r_p_s_redex_reduct:
  assumes step: "(s,t) \<in> qrstep_r_p_s nfs Q R lr p \<sigma>"
  shows "t|_p = snd lr \<cdot> \<sigma>"
    proof - 
    have p_in_poss_t: "p \<in> poss t" using qrstep_r_p_s_imp_poss[OF step] by simp
    have "ctxt_of_pos_term p s = ctxt_of_pos_term p t" (is "_ = ?C") using ctxt_of_pos_term_qrstep_below[OF step] by simp
    from this step[unfolded qrstep_r_p_s_def] have t: "?C\<langle>snd lr \<cdot> \<sigma>\<rangle> = t" by auto
    show ?thesis using subt_at_ctxt_of_pos_term[OF t p_in_poss_t].
qed


locale innermost_wf = 
  fixes D :: "'f sig"
  and R :: "('f, 'v) trs"
  and Q :: "('f, 'v) term set"
  assumes defs_R: "defs R \<subseteq> D"
  and wf_R: "wf_trs R"
  and innermost: "NF_terms Q \<subseteq> NF_trs R"
begin


abbreviation rew :: "('f,'v) term rel"
  where "rew \<equiv> qrstep False Q R"

abbreviation rew_r_p_s :: "('f, 'v) rule \<Rightarrow> pos \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, 'v) term rel"
  where "rew_r_p_s \<equiv> qrstep_r_p_s False Q R"

lemma rew_r_p_s_imp_NF_subst:
  assumes "(s,t) \<in> rew_r_p_s lr p \<sigma>"
  and "x \<in> vars_rule lr"
  shows "\<sigma> x \<in> NF_trs R"
proof - 
  let ?l = "fst lr"
  have "lr \<in> R" using assms(1)[unfolded qrstep_r_p_s_def] by auto
  with assms(2) wf_R[unfolded wf_trs_def'] obtain f ls where 
    l: "?l = Fun f ls" and x: "x \<in> vars_term ?l"
    by (force simp: vars_rule_def)+
  have "\<sigma> x \<in> NF_terms Q" unfolding l using 
    subst_image_subterm[OF x[unfolded l], of \<sigma>] assms(1)[unfolded qrstep_r_p_s_def] l by auto
  thus ?thesis using innermost by auto
qed

abbreviation "step s t \<equiv> (s,t) \<in> rew"
notation step ("(_/ \<rightarrow> _)" [56, 56] 55)

definition DPos :: "('f, 'v) Term.term \<Rightarrow> pos set"
  where "DPos s \<equiv> { p . p \<in> poss s \<and> (\<exists> f \<in> D. root (s|_ p) = Some f)}"

lemma step_at_DPos:
  assumes "(s,t) \<in> rew_r_p_s lr p \<sigma>"
  shows "p \<in> DPos s"
proof -
  obtain l r where [simp]: "lr = (l,r)" by force
  from assms[unfolded qrstep_r_p_s_def] have lr: "(l,r) \<in> R" by auto
  from wf_trs_imp_lhs_Fun[OF wf_R lr] obtain f ls where l: "l = Fun f ls" by blast
  from assms(1)[unfolded qrstep_r_p_s_def] obtain \<sigma> where 
    s_p: "s |_ p = l \<cdot> \<sigma>" by auto
  have rt_s_p: "root (s |_ p) = root l" unfolding l s_p  by auto  
  obtain f where "root (s |_ p) = Some f" and "f \<in> D"
    unfolding rt_s_p
    using lr defs_R[unfolded defs_def defined_def] unfolding l by fastforce
  thus ?thesis unfolding DPos_def l using qrstep_r_p_s_conv(1)[OF assms(1)] by force
qed

definition RPos :: "('f, 'v) Term.term \<Rightarrow> pos set"
  where "RPos s \<equiv> { p . p \<in> DPos s \<and> (\<exists> u. s |_ p \<rightarrow> u)}"

lemma step_at_RPos:
  assumes "(s,t) \<in> rew_r_p_s lr p \<sigma>"
  shows "p \<in> RPos s"
  unfolding RPos_def
  proof (auto)
    from step_at_DPos[OF assms] show p_in_dposs_s: "p \<in> DPos s".
    from qrstep_subt_at[OF assms] have "s |_ p \<rightarrow> t |_ p" 
    using qrstep_qrstep_r_p_s_conv by blast
  thus "\<exists> u. s |_ p \<rightarrow> u" by auto
qed

lemma RPos_subseteq_DPos: "RPos s \<subseteq> DPos s" unfolding RPos_def by auto

lemma DPos_subseteq_funposs:
  "DPos s \<subseteq> funposs s"
proof
  fix p
  assume *: "p \<in> DPos s"
  {
    assume "p \<in> varposs s" 
    then obtain x where "s |_ p = Var x" using varposs_iff by force
    moreover from * obtain f n where "root (s |_ p) = Some (f,n)" unfolding DPos_def by auto
    ultimately have False by auto
  }
  note not_var = this
  from * have "p \<in> (funposs s \<union> varposs s)" unfolding DPos_def by auto
  then show "p \<in> funposs s" using not_var by force
qed

lemma RPos_step_approx: 
  assumes pstep: "(s,t) \<in> rew_r_p_s lr p \<sigma>" 
  shows "RPos t \<subseteq> {q. q \<in> RPos s \<and> (q < p \<or> q \<bottom> p)} \<union> {p <#> q | q. q \<in> DPos (snd lr)}" (is "_ \<subseteq> ?old \<union> ?fresh")
proof
  let ?l = "fst lr"
  let ?r = "snd lr"
  fix q
  assume *: "q \<in> RPos t"
  show "q \<in> ?old \<union> ?fresh"
  proof - 
    have p_in_poss_s: "p \<in> poss s" and p_in_poss_t : "p \<in> poss t" and shape_s: "s = replace_at s p (?l \<cdot> \<sigma>)"
      using qrstep_r_p_s_conv[OF pstep] by auto
    have t_p: "t |_ p = ?r \<cdot> \<sigma>" using subt_at_ctxt_of_pos_term[OF _ p_in_poss_t] qrstep_r_p_s_conv[OF pstep] by auto

    from *[unfolded RPos_def] obtain u where 
      qstep: "t|_q \<rightarrow> u" and
      q_in_DPos_t: "q \<in> DPos t"
      by auto 
    then have q_in_poss_t: "q \<in> poss t" unfolding DPos_def by auto
    from q_in_DPos_t have q_in_funnpos_t: "q \<in> funposs t" using DPos_subseteq_funposs by auto

    {
      assume le: "p \<le> q" 
      obtain p' where p': "q = p <#> p'" using prefix_pos_diff[OF le,symmetric] by auto
      hence 
        t_q: "t |_ q = ?r \<cdot> \<sigma> |_ p'" 
        and p'_in_poss_rsigma: "p' \<in> poss (?r \<cdot> \<sigma>)" 
        using subterm_poss_conv[OF q_in_poss_t p' t_p] by auto

      { 
        assume p'_notin_funposs_r: "p' \<notin> funposs ?r"
        obtain p1 p2 x
          where p'_decomposed: "p' = p1 <#> p2"
          and p1_in_poss_r: "p1 \<in> poss ?r"
          and p1_varposs: "?r |_ p1 = Var x"
          and p2_in_subst: "p2 \<in> poss (\<sigma> x)"
          using poss_subst_apply_term[OF p'_in_poss_rsigma p'_notin_funposs_r].
        have t_q_in_subst: "\<sigma> x \<unrhd> t |_ q"
          unfolding t_q p'_decomposed 
            subt_at_append[OF poss_imp_subst_poss[OF p1_in_poss_r]]
            subt_at_subst[OF p1_in_poss_r]
            p1_varposs
          using subt_at_imp_supteq[OF p2_in_subst] by auto
        have "x \<in> vars_rule lr" 
          using subt_at_imp_supteq[OF p1_in_poss_r, unfolded p1_varposs] var_subt_imp_subt_in_vars 
          unfolding vars_rule_def by force
        then have "(\<sigma> x) \<in> NF_trs R" using rew_r_p_s_imp_NF_subst[OF pstep, of x] by auto
        from NF_subterm[OF this t_q_in_subst] have "t |_q \<in> NF_trs R".
        then have t_q_nf: "t |_ q \<in> NF (qrstep False Q R)" 
          using NF_anti_mono[OF qrstep_subset_rstep] innermost by force
        from NF_no_step[OF t_q_nf] qstep have False by blast
      }
      hence p'_in_funposs_r: "p' \<in> funposs ?r" by auto
      have p'_in_poss_r: "p' \<in> poss ?r" using funposs_imp_poss[OF p'_in_funposs_r].
      from funposs_fun_conv[OF p'_in_funposs_r] obtain f ts where rp': "?r |_ p' = Fun f ts" by auto
      let ?root = "root (?r |_ p')"
      from rp' have "?root = root(?r |_ p' \<cdot> \<sigma>)" unfolding rp' by simp
      also have "... = root(t |_ q)" unfolding t_q using p'_in_poss_r by simp
      finally have rt_r_p': "?root = root(t |_ q)" .
      then obtain f n where "?root = Some (f,n)" and "(f,n) \<in> D" using 
        q_in_DPos_t[unfolded DPos_def] q_in_poss_t by auto
      then have "p' \<in> DPos ?r" unfolding DPos_def using p'_in_poss_r by auto
      then have ?thesis using p' by auto
    }
    note q_geq_p = this

    {
      assume p_gt_q: "p > q"
      then obtain p' where p' : "p = q <#> p'" and ne: "p' \<noteq> \<epsilon>" using less_pos_def' by auto
      have step_at_q: "(s |_ q, t |_ q) \<in> nrqrstep False Q R"
        using qrstep_r_p_s_imp_nrqrstep[OF qrstep_subt_at_gen[OF pstep[unfolded p']] ne] .
      from nrqrstep_preserves_root[OF this]
      obtain f n where 
        "root (s |_ q) = Some (f,n)" and
        "(f,n) \<in> D"
        using q_in_poss_t q_in_DPos_t[unfolded DPos_def] by auto
      hence "q \<in> DPos s" unfolding DPos_def 
        using poss_append_poss p_in_poss_s[unfolded p'] by auto
      then have "q \<in> ?old" using step_at_q 
        unfolding RPos_def
        using qrstep_iff_rqrstep_or_nrqrstep p_gt_q
        by blast
      then have ?thesis by auto
    } 
    note p_gt_q = this
    
    {
      assume p_par_q: "p \<bottom> q"
      have q_in_poss_s: "q \<in> poss s" unfolding shape_s[unfolded qrstep_r_p_s_conv(6)[OF pstep, symmetric]] 
        using parallel_poss_replace_at[OF p_par_q p_in_poss_t] q_in_poss_t by auto
      have "s |_ q = t |_ q" using parallel_qrstep_subt_at[OF pstep p_par_q q_in_poss_s] by auto
      then obtain f n where 
        "root(s |_ q) = Some (f,n)"
        and "(f,n) \<in> D"
        and step_q: "s |_ q \<rightarrow> u"
        using *[unfolded RPos_def DPos_def] qstep by auto
      then have "q \<in> DPos s"using q_in_poss_s unfolding DPos_def by blast
      then have ?thesis using step_q parallel_pos_sym[OF p_par_q] unfolding RPos_def by blast
    }
    note q_par_q = this
    show ?thesis using q_geq_p p_gt_q q_par_q position_cases by auto
  qed
qed
end


locale sorryy =
  assumes sorryy: False


lemma (in sorryy) fill_holes_mctxt_sound:
  assumes "num_holes C = length Ds"
  and "length sss = length Cs"
  and "\<forall> i < length Cs. num_holes (Ds!i) = length (sss!i)"
  shows "num_holes (fill_holes_mctxt C Ds) = listsum (map num_holes Ds)"
  and  "fill_holes (fill_holes_mctxt C Ds) (concat sss) = fill_holes C [fill_holes (Cs!i) (sss!i) . i \<leftarrow> [0..<length Cs] ]"
 using sorryy by auto
(*  
proof (induct C  arbitrary: Ds)

case (MFun f Cs) 
  let ?Ds = "\<lambda> i. take (num_holes (Cs ! i)) (drop (listsum (map num_holes (take i Cs))) Ds)" 
  let ?f = "\<lambda> i. fill_holes_mctxt (Cs ! i) (?Ds i)"

  have Ds: "concat [?Ds i . i \<leftarrow> [0..< length Cs]] = Ds"  sorry
      
(*  {
    fix i
    assume i: "i < length Cs"
    have "num_holes (?f i) = listsum (map num_holes (?Ds i))" using MFun(1)[OF nth_mem[OF i]] .
  } note num_holes_i = this *)


  have "num_holes (fill_holes_mctxt (MFun f Cs) Ds) = listsum (map num_holes [?f i . i \<leftarrow> [0 ..< length Cs]])" by simp
  moreover have "... = listsum [ (num_holes \<circ> ?f) i . i \<leftarrow> [0 ..< length Cs]]" by (metis map_map o_def)
  moreover have "... = listsum [ num_holes (?f i) . i \<leftarrow> [0 ..< length Cs]]" by (metis o_def)
  moreover have "... = listsum [ listsum (map num_holes (?Ds i)) . i \<leftarrow> [0 ..< length Cs]]" 
    by (smt MFun.hyps in_set_conv_nth length_map map_eq_conv map_nth nth_map)
*)

fun
  term_to_mctxt :: "('f, 'v) term \<Rightarrow> ('f, 'v) mctxt" where 
   "term_to_mctxt (Var x) = (MVar x)"
  | "term_to_mctxt (Fun f ts) = MFun f [ term_to_mctxt t . t \<leftarrow> ts]"

lemma term_to_mctxt_numholes:
  "num_holes (term_to_mctxt t) = 0"
proof (induct t, auto) qed

lemma term_to_mctxt_sound:
  "fill_holes (term_to_mctxt t) [] = t"
  proof (induct t)
    case (Fun f ts) 
      let ?Cs = "[ term_to_mctxt ti . ti \<leftarrow> ts]"
      let ?ss = "[ [] . ti \<leftarrow> ts]"
      have l_Cs: "length ?Cs = length ts" and l_ss: "length ?ss = length ts" by auto
      have concat_ss: "concat ?ss = []" by auto
      {
        fix i
        assume i: "i < length ts"
        have "num_holes (?Cs ! i) = length (?ss ! i)" and
          "fill_holes (?Cs ! i) (?ss ! i) = ts ! i" 
          unfolding nth_map[OF i] list.size using Fun(1)[OF nth_mem[OF i]] term_to_mctxt_numholes by auto
      } hence fill_holes_i: "\<And> i. i < length ts \<Longrightarrow> num_holes (?Cs ! i) = length (?ss ! i) \<and> fill_holes (?Cs ! i) (?ss ! i) = ts ! i" by auto
      
      show ?case unfolding term_to_mctxt.simps fill_holes.simps 
        using fill_holes_arbitrary[OF l_Cs l_ss fill_holes_i, unfolded concat_ss] by metis
    qed auto

      
fun
  subst_apply_mctxt :: "('f, 'v) mctxt \<Rightarrow> ('f, 'v, 'w) gsubst \<Rightarrow> ('f, 'w) mctxt"
  (infixl "\<cdot>mc" 67) where
  "MHole \<cdot>mc _ = MHole"
|  "(MVar x) \<cdot>mc \<sigma> = term_to_mctxt (\<sigma> x)"
| "(MFun f cs) \<cdot>mc \<sigma> = MFun f [c \<cdot>mc \<sigma> . c \<leftarrow> cs]"


definition ith_hole_terms :: "('a, 'b) term list \<Rightarrow> ('a, 'c) mctxt list \<Rightarrow> nat \<Rightarrow> ('a, 'b) term list" where
  "ith_hole_terms ts cs i \<equiv> take (num_holes (cs ! i)) (drop (listsum (map num_holes (take i cs))) ts)"

lemma ith_hole_terms_len:
  assumes "num_holes (MFun f Cs) = length ts" and "i < length Cs"
  shows "num_holes (Cs!i) = length (ith_hole_terms ts Cs i)"
proof -
  have "Cs!i = hd (drop i Cs)" using hd_drop_conv_nth[OF assms(2)] by auto
  moreover have "drop i Cs \<noteq> []" using drop_eq_Nil assms(2) using not_le by auto
  ultimately have "(Cs!i) \<in> set (drop i Cs)" using hd_in_set by auto
  hence "num_holes (Cs!i) \<in> set (map num_holes (drop i Cs))" by auto
  hence *: "num_holes (Cs!i) \<le> listsum (map num_holes (drop i Cs))" using member_le_listsum_nat by auto

  have "length ts = listsum (map num_holes (take i Cs) @ map num_holes (drop i Cs))"
    unfolding assms(1)[symmetric] num_holes.simps
    map_append[symmetric] append_take_drop_id by simp
  moreover have "... = listsum (map num_holes (take i Cs)) + listsum (map num_holes (drop i Cs))" using listsum_append by auto
  ultimately have "length (drop (listsum (map num_holes (take i Cs))) ts) = listsum (map num_holes (drop i Cs))"
    by (metis add_diff_cancel_left' length_drop) 
  from this * show ?thesis unfolding ith_hole_terms_def using length_take by (metis min_absorb2)
qed


lemma (in sorryy) concat_ith_hole_terms:
  assumes "t=f (MFun f Cs, ts)"
  shows "concat (map (\<lambda> i. ith_hole_terms ts Cs i) [0..<length Cs]) = ts"
  using sorryy ..

lemma (in sorryy) subst_apply_mxtct_sound: 
  "t =f (C,ts) \<Longrightarrow> t \<cdot> \<sigma> =f (C \<cdot>mc \<sigma>, [ ti \<cdot> \<sigma> . ti \<leftarrow> ts])"
  using sorryy ..
(*
proof (induct C arbitrary: t ts)
  case MHole 
  note nh = eqfE(2)[OF MHole]
  have lts: "length [t] = length ts" using  nh unfolding num_holes.simps  by auto
  from eqfE(1)[OF MHole] fill_holes.simps(2)[of t] have fh: "fill_holes MHole ts = fill_holes MHole [t]" by auto
  have "ts = [t]" using fill_holes_inj[OF fh nh lts] .
  then show ?case by auto

  next
  case (MVar x) 
  note nh = eqfE(2)[OF MVar]
  have ts: "ts = []" by (metis length_0_conv nh num_holes.simps(1))
  have t: "t = Var x" using eqfE(1)[OF MVar, unfolded fill_holes.simps] .
  show ?case unfolding t ts subst_apply_mctxt.simps subst_apply_term.simps proof (rule eqfI, insert term_to_mctxt_numholes term_to_mctxt_sound[symmetric], auto) qed
  next
  case (MFun f Cs) show ?case proof (auto)
    let ?us = "[fill_holes (Cs ! i) (ith_hole_terms ts  Cs i) . i \<leftarrow> [0..<length Cs]]"
    note IH = MFun(1)
    from eqfE[OF MFun(2)] have 
      nh: "num_holes (MFun f Cs) = length ts" and
      fh: "fill_holes (MFun f Cs) ts = t" by auto
    then have t: "t = Fun f ?us" unfolding ith_hole_terms_def by auto

    let ?us' = "[ ui \<cdot> \<sigma> . ui \<leftarrow> ?us]"
    let ?ts' = "[ ti \<cdot> \<sigma> . ti \<leftarrow> ts]"
    let ?Cs' = "map (\<lambda> c . c \<cdot>mc \<sigma>) Cs"
    let ?tss = "map (\<lambda> i. map (\<lambda>ti. ti \<cdot> \<sigma>) (ith_hole_terms ts Cs i)) [0..<length Cs]"
    {
      fix i
      assume i: "i < length ?us" 
      hence i_Cs: "i < length Cs" using length_map by auto
      from i have i_us': "i < length ?us'" using length_map by auto
      from i have i_Cs': "i < length ?Cs'" using length_map by auto
      from i_Cs have i_lcs: "i < length [0..<length Cs]" using length_map by auto
      then have i_lcs_simp: "[0..<length Cs] ! i = i" by auto
      from i_Cs have i_lcs_simps2: "\<And> fn . [ fn j . j \<leftarrow> [0..<length Cs]] ! i = fn i" by auto
      have iCs: "Cs ! i \<in> set Cs" using nth_mem i by auto 
      have us_i: "?us!i =f (Cs!i, ith_hole_terms ts Cs i)" using ith_hole_terms_len[OF nh] i eqfI by auto

      from eqfE[OF IH[OF iCs us_i]] have 
        "num_holes (?Cs' ! i) = length (?tss ! i)" and
        "fill_holes (?Cs' ! i) (?tss ! i) = ?us' ! i" 
        unfolding nth_map[OF i_Cs] nth_map[OF i_lcs] map_map i_lcs_simp by auto
    } hence fill_holes_i: "\<And> i. i < length ?us' \<Longrightarrow> num_holes (?Cs' ! i) = length (?tss ! i) \<and> fill_holes (?Cs' ! i) (?tss ! i) = ?us' ! i" by auto

    have l_cs'_us': "length ?Cs' = length ?us'" by auto
    have l_tss_us': "length ?tss = length ?us'" by auto
    let ?C' = "MFun f ?Cs'"
    from fill_holes_arbitrary[OF l_cs'_us' l_tss_us' fill_holes_i]
    have "t \<cdot> \<sigma> = fill_holes ?C' (concat ?tss)" unfolding t ith_hole_terms_def fill_holes.simps by auto
    (* moreover from concat_ith_hole_terms[OF MFun(2)] have "... = fill_holes ?C' ts"
    have "concat ?tss = map (\<lambda>ti. ti \<cdot> \<sigma>) ts" *)
*)      

fun compose_mctxt :: "('f,'v)mctxt \<Rightarrow> nat \<Rightarrow> ('f,'v)mctxt \<Rightarrow> ('f,'v)mctxt" where
  "compose_mctxt C i Ci = fill_holes_mctxt C [ if i = j then Ci else MHole . j \<leftarrow> [0..<num_holes C]]"

lemma (in sorryy) compose_mctxt_sound:
  assumes "s =f (C, bef @ si # aft)"
  and "si =f (D, ts)"
  and "i = length bef"
  shows "s =f (compose_mctxt C i D, bef @ ts @ aft)"
  using sorryy ..


lemma (in sorryy) mctxt_fill_partially: 
  assumes "t =f (C,ts)" and "sublist ss ts"
  obtains D where "t =f (D,ss)"
  using sorryy ..

locale DT_trans = innermost_wf D + sorryy for D :: "'f sig" +
  fixes F :: "'f sig"
  and shp :: "'f \<Rightarrow> 'f"
  and Ds :: "'f sig"
  and Cp :: "'f sig"

  defines [simp]: "Ds \<equiv> {(shp f,n) | f n . (f,n) \<in> D}"

  assumes R_sig: "funas_trs R \<subseteq> F"
  and D_subseteq_F: "D \<subseteq> F"
  and Cp_fresh: "Cp \<inter> F = {}"
  and Ds_fresh: "Ds \<inter> F = {}"
begin

fun wf_marked_term :: "('f, 'v) term \<Rightarrow> bool"
  where 
    "wf_marked_term (Var _) = False"
  | "wf_marked_term (Fun f ts) = ((f, length ts) \<in> Ds \<and> (\<forall> ti \<in> set ts. funas_term ti \<inter> (Ds \<union> Cp) = {}))"


fun wf_cpctx_term :: "('f, 'v) term \<Rightarrow> bool"
  where 
    "wf_cpctx_term (Var _) = False" 
  | "wf_cpctx_term (Fun f ts) = ((((f, length ts) \<in> Cp) \<and> (\<forall> ti \<in> set ts. wf_cpctx_term ti)) \<or> wf_marked_term (Fun f ts))"


(* Tuple Transformation *)


definition wf_DT :: "('f, 'a) rule \<Rightarrow> bool"
  where "wf_DT uv \<equiv> wf_marked_term (fst uv) \<and> wf_cpctx_term (snd uv)"
            
  

definition is_DT_of :: "('f, 'a) rule \<Rightarrow> ('f, 'a) rule \<Rightarrow> bool" where 
  "is_DT_of lr uv \<equiv>
    fst uv = sharp_term shp (fst lr) \<and> (\<exists> C qs. snd uv =f (C,[sharp_term shp ((snd lr) |_ q) . q \<leftarrow> qs]) \<and> set qs = DPos (snd lr))"
              

definition goodFor :: "('f, 'a) term  \<Rightarrow> ('f, 'a) term \<Rightarrow> bool" where
  "goodFor t u \<equiv> \<exists> C ps. u =f (C, [sharp_term shp (t |_ p) . p \<leftarrow> ps]) \<and> (RPos t = set ps)"

notation goodFor ("(_/ \<lless> _)" [56,56] 55)

lemma nrqrstep_r_p_s_imp_sharp_qrstep_r_p_s: 
  assumes ne: "p \<noteq> \<epsilon>"
  and step: "(s,t) \<in> rew_r_p_s lr p \<sigma>"
  shows "(sharp_term shp s,sharp_term shp t) \<in> rew_r_p_s lr p \<sigma>"
proof -
  let ?C = "ctxt_of_pos_term p s"
  from qrstep_r_p_s_conv[OF step] have s: "s = ?C\<langle>fst lr\<cdot>\<sigma>\<rangle>" and t: "t = ?C\<langle>snd lr\<cdot>\<sigma>\<rangle>" and p: "p \<in> poss s" by auto
  from ne obtain f D bef aft where C: "?C = More f bef D aft"
    using hole_pos_ctxt_of_pos_term[OF p] ctxt_to_term_list.cases hole_pos.simps(1) by metis
  from s t have s: "s = Fun f (bef @ D\<langle>fst lr\<cdot>\<sigma>\<rangle>  # aft)" and t: "t = Fun f (bef @ D\<langle>snd lr\<cdot>\<sigma>\<rangle> # aft)" unfolding C ctxt_apply.simps(2) by auto
  let ?C' = "More (shp f) bef D aft"
  
  from hole_pos_ctxt_of_pos_term[OF p, unfolded C] have hole_pos_C': "hole_pos ?C' = p" by simp
  from s t have 
    s_sharp: "sharp_term shp s = ?C'\<langle>fst lr\<cdot>\<sigma>\<rangle>" and 
    t_sharp: "sharp_term shp t = ?C'\<langle>snd lr\<cdot>\<sigma>\<rangle>" by simp+
  moreover have "ctxt_of_pos_term p (sharp_term shp s) = ?C'" using s_sharp hole_pos_C' by fastforce
  moreover have "sharp_term shp s |_ p = fst lr \<cdot> \<sigma>" and "sharp_term shp t |_ p = snd lr \<cdot> \<sigma>"
    unfolding s_sharp t_sharp using hole_pos_C' subt_at_hole_pos by auto
  moreover from hole_pos_C' have "p \<in> poss (sharp_term shp s)" unfolding s_sharp using hole_pos_in_poss by auto
  ultimately show ?thesis using step unfolding qrstep_r_p_s_def  by auto 
qed


lemma step_to_dt_step:
  assumes step: "(s,t) \<in> rew_r_p_s lr p \<sigma>"
  and dt: "is_DT_of lr uv"
  obtains C t' qs 
  where "t' =f(C,[sharp_term shp (t|_(p <#> q)) . q \<leftarrow> qs])" 
  and "set qs = DPos (snd lr)"
  and "(sharp_term shp (s |_ p), t') \<in> qrstep False Q {uv}"
  using sorryy ..


lemma 
  assumes step: "(s,t) \<in> rew_r_p_s lr p \<sigma>"
  and dt: "is_DT_of lr lr'"
  and good: "s \<lless> u"
  shows "\<exists> v. t \<lless> v \<and> (u,v) \<in> ( rew^* O qrstep False Q {lr'} )"
proof -
  let ?sharp_at = "\<lambda> tm q. sharp_term shp (tm |_ q)"

  from good[unfolded goodFor_def] obtain C rpos_s_list where 
    u_decompose: "u =f (C,[?sharp_at s q . q \<leftarrow> rpos_s_list])" and
    rpos_s_list: "RPos s = set rpos_s_list"
    by auto

  let ?f = "\<lambda> q. if q < p \<or> q \<bottom> p then ?sharp_at t q else ?sharp_at s q"
  let ?us = "[ ?sharp_at s q . q \<leftarrow> rpos_s_list ]"
  let ?ws = "[ ?f q . q \<leftarrow> rpos_s_list ]"
  
  let ?w = "fill_holes C ?ws"
  {
    fix q
    assume q_in_rpos_s: "q \<in> set rpos_s_list"
    { 
      assume q: "q < p"
      then have q_diff: "p = q <#> pos_diff p q" (is "_ = _ <#> ?q'") by auto 
      then have ne: "?q' \<noteq> \<epsilon>" using q less_simps(1) by force
      have "(s |_ q, t |_ q) \<in> rew_r_p_s lr ?q' \<sigma>" by (metis q_diff qrstep_subt_at_gen step)
      from nrqrstep_r_p_s_imp_sharp_qrstep_r_p_s[OF ne this] have
        "(?sharp_at s q, ?f q) \<in> rew_r_p_s lr ?q' \<sigma>" using q by auto
      hence "(?sharp_at s q, ?f q) \<in> rew" using qrstep_qrstep_r_p_s_conv by blast
      hence "(?sharp_at s q, ?f q) \<in> rew^*" by auto
    } note q_less_p = this

    { 
      assume q_par_p: "q \<bottom> p"
      from q_in_rpos_s have "q \<in> poss s" 
        unfolding rpos_s_list[symmetric] RPos_def DPos_def by auto
      from parallel_qrstep_subt_at[OF step _ this] have "?f q = ?sharp_at s q" 
        using q_par_p parallel_pos_sym by auto
      then have "(?sharp_at s q, ?f q) \<in> rew^*" by auto
    } note q_par_p_p = this

    { 
      assume q: "\<not> (q < p \<or> q \<bottom> p)"
      then have "?f q = ?sharp_at s q" by auto
      then have "(?sharp_at s q, ?f q) \<in> rew^*" by auto
    } note q_nless_p = this

    have "(?sharp_at s q, ?f q) \<in> rew^*" using q_par_p_p q_less_p q_nless_p by auto
  } note repair_step = this
  
  { 
    fix j
    assume "j < length ?ws"
    then have "(?us!j,?ws!j) \<in> rew^*" using repair_step rpos_s_list by auto
  } 
  
  from eqf_all_ctxt_closed_step[OF all_ctxt_closed_qrsteps u_decompose this] length_map 
  have repair_steps: "(u,?w) \<in> (rew^*)" by auto

  obtain Dp t' dpos_r_list where
    t': "t' =f (Dp, [?sharp_at t (p <#> q) . q \<leftarrow> dpos_r_list])"
    and dpos_r_list_dpos: "set dpos_r_list = DPos (snd lr)"
    and p_step: "(?sharp_at s p, t') \<in> qrstep False Q {lr'}"
    using step_to_dt_step[OF step dt] by metis

  let ?rs = "[?sharp_at t (p <#> q) . q \<leftarrow> dpos_r_list]"
  obtain i where 
    rpos_s_list_i: "rpos_s_list!i = p" and 
    i_index_rpos_s_list: "i < length rpos_s_list" 
    using rpos_s_list step_at_RPos[OF step] List.in_set_conv_nth[of p rpos_s_list] by auto
  let ?rpos_s_left = "take i rpos_s_list"
  let ?rpos_s_right = "drop (Suc i) rpos_s_list"


  have i_index_ws: "i < length ?ws" using i_index_rpos_s_list by auto
  have nh_C_l_ws: "num_holes C = length ?ws" using eqfE[OF u_decompose] length_map by auto

  from fill_holes_ctxt[OF this i_index_ws] obtain C' where 
    C': "\<And> s. fill_holes C (?ws[i := s]) = C' \<langle> s \<rangle>" by auto
  
  let ?v = "fill_holes C (?ws[i := t'])"
  have ws_i: "?ws!i = ?sharp_at s p" 
    unfolding nth_map[OF i_index_rpos_s_list] rpos_s_list_i  
    using less_irrefl parallel_pos order_refl by auto
  then have "?ws = ?ws[i := ?sharp_at s p]" unfolding ws_i[symmetric] using  list_update_id[of ?ws i] by auto
  from this C' have C'ws: "?w = C'\<langle>?sharp_at s p\<rangle>" by (metis (lifting))
  from C' have C'vs: "?v = C'\<langle>t'\<rangle>" by (metis (lifting))
  
  have dp_step: "(fill_holes C ?ws, ?v) \<in> qrstep False Q {lr'}" unfolding C'ws C'vs using qrstep_closed_ctxt[OF p_step] .

  from repair_steps dp_step have step_simulation: "(u,?v) \<in> ( rew^* O qrstep False Q {lr'} )" by auto

  let ?ws_left = "[ ?f q . q \<leftarrow> ?rpos_s_left]"
  let ?ws_right = "[ ?f q . q \<leftarrow> ?rpos_s_right]"
  have sep: "?ws_left @ t' # ?ws_right = ?ws[i := t']" 
    unfolding take_map[symmetric] drop_map[symmetric]
    using id_take_nth_drop[OF i_index_ws] upd_conv_take_nth_drop[OF i_index_ws] by auto
  let ?vs = "?ws_left @ ?rs @ ?ws_right"
  let ?D = "compose_mctxt C i Dp"

  have len_left: "length ?ws_left = i" using length_take i_index_rpos_s_list by auto
  moreover have "?v =f (C,?ws_left @ t' # ?ws_right)" unfolding sep using nh_C_l_ws by auto
  ultimately have v_decompose_D: "?v =f (?D, ?vs)" 
    using compose_mctxt_sound[OF _ t'] by (metis (lifting, no_types))


  let ?filter_old = "filter (\<lambda> q. q \<in> RPos s \<and> (q < p \<or> q \<bottom> p))"
  let ?qs = "(?filter_old ?rpos_s_left) @ [p <#> q . q \<leftarrow> dpos_r_list] @ (?filter_old ?rpos_s_right)"
  let ?rpos_t_list = "[q \<leftarrow> ?qs. q \<in> RPos t]"
  have good_RPos: "RPos t = set ?rpos_t_list" proof
    {
      fix q
      assume q_in_rpos_t: "q \<in> RPos t"
      then have q_in_approx: "q \<in> {q. q \<in> RPos s \<and> (q < p \<or> q \<bottom> p)} \<or> q \<in> {p <#> q | q. q \<in> DPos (snd lr)}" (is "_ \<in> ?old \<or> _ \<in> ?new") 
        using RPos_step_approx[OF step] by auto
      {
        assume q_in_old: "q \<in> ?old"
        hence "q < p \<or> q \<bottom> p" by auto hence q_noteq_p: "q \<noteq> p" using less_pos_def parallel_pos by auto

        have "q \<in> set rpos_s_list" using q_in_old rpos_s_list by auto 
        hence "q \<in> set (?rpos_s_left @ rpos_s_list!i # ?rpos_s_right)" 
          by (metis i_index_rpos_s_list id_take_nth_drop)
        hence "q \<in> set (?rpos_s_left @ ?rpos_s_right)" 
          unfolding rpos_s_list_i using q_noteq_p by auto
        hence "q \<in> set ?qs" using q_in_old by auto
      } note q_in_old = this
      { 
        assume q_in_new: "q \<in> ?new"
        hence "q \<in> set ?qs" unfolding dpos_r_list_dpos[symmetric] by auto
      } note q_in_new = this
      
      from q_in_approx q_in_old q_in_new q_in_rpos_t have "q \<in> set ?rpos_t_list" by auto
    }
    then show "RPos t \<subseteq> set ?rpos_t_list" by auto
    next show "set ?rpos_t_list \<subseteq> RPos t" by auto
  qed

  { fix q 
    fix pos_lst
    assume "q \<in> set (?filter_old pos_lst)"
    hence "q < p \<or> q \<bottom> p" by auto hence "?sharp_at t q = ?f q" by simp
  } 
  hence f_of_q_in_filter_old: "\<And> pos_lst. \<forall> q \<in> set (?filter_old pos_lst). ?sharp_at t q = ?f q" by blast

  have good_sharp: "sublist [?sharp_at t q . q \<leftarrow> ?rpos_t_list] ?vs"
    proof -
      have left: "sublist [ ?sharp_at t q. q \<leftarrow> ?filter_old ?rpos_s_left] ?ws_left" (is "sublist ?lft _")
        using map_sublist_f_g[OF filter_sublist f_of_q_in_filter_old] .
      
      have right: "sublist [ ?sharp_at t q. q \<leftarrow> ?filter_old ?rpos_s_right] ?ws_right" (is "sublist ?rght _")
        using map_sublist_f_g[OF filter_sublist f_of_q_in_filter_old] .

      have middle: "sublist [ ?sharp_at t q . q \<leftarrow> [(p <#> q'). q' \<leftarrow> dpos_r_list]] ?rs" (is "sublist ?mid _") 
        unfolding map_map  comp_def using sublist_refl by auto

      have "[?sharp_at t q. q \<leftarrow> ?qs ] = ?lft @ ?mid @ ?rght" 
        unfolding map_append by simp
      hence "sublist [?sharp_at t q. q \<leftarrow> ?qs ] ?vs" using left middle right sublist_append
        by (metis (lifting, no_types))
      thus ?thesis using filter_sublist map_sublist sublist_trans by blast    
    qed


  from mctxt_fill_partially[OF v_decompose_D good_sharp] obtain C' where
    "?v =f (C',[?sharp_at t q . q \<leftarrow> ?rpos_t_list])" by auto
  from v_decompose_D good_RPos this
  have good: "t \<lless> ?v" unfolding goodFor_def
    by (metis (lifting, no_types))
  then show ?thesis using step_simulation by auto
qed

  
end
  



  
  

