theory Renaming
imports
  Trs
  Fresh
  Option_Monad
begin

definition vsubst :: "('f, 'v) subst \<Rightarrow> bool" where
  "vsubst \<sigma> \<longleftrightarrow> (\<forall>x. is_Var (\<sigma> x))"

definition vren :: "('f, 'v) subst \<Rightarrow> bool" where
  "vren \<sigma> \<longleftrightarrow> vsubst \<sigma> \<and> bij (the_Var \<circ> \<sigma>)"

lemma vren_imp_inj:
  "vren \<sigma> \<Longrightarrow> inj (the_Var \<circ> \<sigma>)"
  by (auto simp: vren_def bij_betw_def)

lemma the_Var_Var [simp]:
  "the_Var \<circ> Var = (\<lambda>x. x)"
  by (rule ext) auto

lemma vren_Var [simp]:
  "vren Var"
  by (auto simp: vren_def vsubst_def bij_def)

definition swap :: "'v \<Rightarrow> 'v \<Rightarrow> ('f, 'v) subst" where
  "swap x y = (\<lambda>z. if z = x then Var y else if z = y then Var x else Var z)"

lemma swap_simps [simp]:
  "swap x y x = Var y"
  "swap x y y = Var x"
  "z \<noteq> x \<Longrightarrow> z \<noteq> y \<Longrightarrow> swap x y z = Var z"
  by (auto simp: swap_def)

lemma is_Var_swap [simp]:
  "is_Var (swap x y z)"
  by (cases "x = z", auto) (cases "y = z", auto)

lemma swap_inj:
  "swap x y u = swap x y v \<Longrightarrow> u = v"
  by (auto simp: swap_def [abs_def])
     (metis the_Var.simps)

lemma the_Var_swap_inj:
  "the_Var (swap x y u) = the_Var (swap x y v) \<Longrightarrow> u = v"
  by (metis swap_def the_Var.simps)

lemma range_swap_range_Var_conv:
  "range (swap x y) = (range Var :: ('f, 'v) term set)"
proof
  show "range (swap x y) \<subseteq> (range Var :: ('f, 'v) term set)"
  proof
    fix z :: "('f, 'v) term"
    assume "z \<in> range (swap x y)"
    then obtain u where [simp]: "z = Var u" by (auto simp: swap_def [abs_def])
    then show "z \<in> Var ` UNIV" by auto
  qed
next
  show "(range Var :: ('f, 'v) term set) \<subseteq> range (swap x y)"
  proof
    fix z :: "('f, 'v) term"
    assume "z \<in> range (Var :: 'v \<Rightarrow> ('f, 'v) term)"
    then obtain u where [simp]: "z = Var u" by auto
    have "swap x y (if u = x then y else if u = y then x else u) = z" by auto
    then show "z \<in> range (swap x y)" by blast
  qed
qed

lemma bij_swap:
  "bij (the_Var \<circ> swap x y)"
  by (auto simp: bij_def inj_on_def intro: the_Var_swap_inj)
     (metis comp_apply image_compose rangeI range_swap_range_Var_conv the_Var.simps)

lemma vren_subst [simp]:
  "vren (swap x y)"
  by (auto simp: vren_def bij_swap vsubst_def)

definition vren_inv :: "('f, 'v) subst \<Rightarrow> ('f, 'v) subst" where
  "vren_inv \<sigma> = Var \<circ> inv (the_Var \<circ> \<sigma>)"

lemma subst_compose_vren_inv [simp]:
  assumes "vren \<sigma>"
  shows "\<sigma> \<circ>s vren_inv \<sigma> = Var"
proof (rule ext)
  fix x
  from assms obtain z where [simp]: "\<sigma> x = Var z" by (auto simp add: vren_def vsubst_def)
  from assms have "inj (the_Var \<circ> \<sigma>)" by (rule vren_imp_inj)
  from this [unfolded inj_iff]
    have "(inv (the_Var \<circ> \<sigma>) \<circ> (the_Var \<circ> \<sigma>)) x = x" by simp
  then show "(\<sigma> \<circ>s vren_inv \<sigma>) x = Var x"
    by (simp add: subst_compose_def vren_inv_def)
qed

lemma subst_apply_vren_inv [simp]:
  assumes "vren \<sigma>"
  shows "t \<cdot> \<sigma> \<cdot> vren_inv \<sigma> = t"
  using subst_compose_vren_inv [OF assms]
  by (metis subst_apply_term_empty subst_compose_apply_term_distrib)

lemma vren_inv_subst_compose [simp]:
  assumes "vren \<sigma>"
  shows "vren_inv \<sigma> \<circ>s \<sigma> = Var"
proof (rule ext)
  fix x
  from assms obtain z where [simp]: "\<sigma> x = Var z" by (auto simp add: vren_def vsubst_def)
  from assms have "surj (the_Var \<circ> \<sigma>)" by (simp add: vren_def bij_betw_def)
  from this [unfolded surj_iff]
    have "(the_Var \<circ> \<sigma> \<circ> inv (the_Var \<circ> \<sigma>)) x = x" by simp
  then have "(\<sigma> \<circ> inv (the_Var \<circ> \<sigma>)) x = Var x"
    using assms
    by (auto simp add: o_def vren_def vsubst_def)
       (metis Var_the_Var_ident)
  then show "(vren_inv \<sigma> \<circ>s \<sigma>) x = Var x"
    by (simp add: subst_compose_def vren_inv_def)
qed

lemma subst_apply_vren_inv' [simp]:
  assumes "vren \<sigma>"
  shows "t \<cdot> vren_inv \<sigma> \<cdot> \<sigma> = t"
  using vren_inv_subst_compose [OF assms]
  by (metis subst_apply_term_empty subst_compose_apply_term_distrib)

definition variant :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool" (infix "=v=" 55) where
  "s =v= t \<longleftrightarrow> (\<exists>\<sigma>. vren \<sigma> \<and> s = t \<cdot> \<sigma>)"

lemma variant_refl [simp]:
  "s =v= s"
proof -
  have "s = s \<cdot> Var" by simp
  moreover have "vren Var" by simp
  ultimately show ?thesis
    unfolding variant_def by blast
qed

lemma vren_vren_inv:
  assumes "vren \<sigma>"
  shows "vren (vren_inv \<sigma>)"
proof (unfold vren_def, rule)
  show "vsubst (vren_inv \<sigma>)" by (simp add: vsubst_def vren_inv_def)
  have "bij (the_Var \<circ> Var)"
    by (auto simp: bij_betw_def inj_on_def)
  moreover have "bij (inv (the_Var \<circ> \<sigma>))"
    using assms by (auto simp: vren_def bij_imp_bij_inv)
  ultimately have "bij ((the_Var \<circ> Var) \<circ> (inv (the_Var \<circ> \<sigma>)))" by (metis bij_comp)
  then show "bij (the_Var \<circ> vren_inv \<sigma>)"
    by (auto simp: vren_inv_def o_assoc)
qed

lemma variant_sym:
  assumes "s =v= t"
  shows "t =v= s"
proof -
  from assms obtain \<sigma> where "vren \<sigma>"
    and "s = t \<cdot> \<sigma>" by (auto simp: variant_def)
  then have "s \<cdot> vren_inv \<sigma> = t \<cdot> (\<sigma> \<circ>s vren_inv \<sigma>)" by (metis subst_compose_apply_term_distrib)
  with `vren \<sigma>` have "t = s \<cdot> vren_inv \<sigma>" by simp
  moreover have "vren (vren_inv \<sigma>)" by (rule vren_vren_inv [OF `vren \<sigma>`])
  ultimately show ?thesis
    unfolding variant_def by blast
qed

lemma vsubst_compose [simp]:
  assumes "vsubst \<sigma>"
  shows "\<sigma> \<circ>s \<tau> = \<tau> \<circ> (the_Var \<circ> \<sigma>)"
  using assms
  by (auto simp: vsubst_def subst_compose_def o_def)
     (metis Var_the_Var_ident subst_apply_term.simps(1))

lemma vren_compose [simp]:
  assumes "vren \<sigma>" and "vren \<tau>"
  shows "vren (\<sigma> \<circ>s \<tau>)"
proof -
  from assms have "vsubst \<sigma>" and "vsubst \<tau>" by (auto simp: vren_def)
  then have "vsubst (\<sigma> \<circ>s \<tau>)"
    by (auto simp add: vsubst_def subst_compose_def)
       (metis is_VarE subst_apply_term.simps(1))
  moreover
  {
    from assms have "bij (the_Var \<circ> \<sigma>)" and "bij (the_Var \<circ> \<tau>)" by (auto simp: vren_def)
    then have "bij (the_Var \<circ> (\<sigma> \<circ>s \<tau>))"
      using `vsubst \<sigma>`
      by (auto simp: o_assoc)
         (metis bij_def comp_inj_on_iff image_compose inj_on_imageI2)
  }
  ultimately show ?thesis by (auto simp: vren_def)
qed

lemma variant_trans:
  assumes "s =v= t" and "t =v= u" shows "s =v= u"
proof -
  from assms obtain \<sigma> \<tau>
    where "vren \<sigma>" and "vren \<tau>"
    and "s = t \<cdot> \<sigma>" and "t = u \<cdot> \<tau>" by (auto simp: variant_def)
  then have "s = u \<cdot> (\<tau> \<circ>s \<sigma>)" and "vren (\<tau> \<circ>s \<sigma>)" by simp+
  then show ?thesis by (unfold variant_def) blast
qed

text {*@{const variant} is an equivalence relation on terms.*}
lemma equivp_variant: "equivp (op =v=)"
  by (rule equivpI)
     (auto simp: reflp_def symp_def transp_def intro: variant_sym variant_trans)

subsection {* Variable-Disjoint Variants *}

definition disj_variant :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool" (infix "<v>" 55) where
  "s <v> t \<longleftrightarrow> s =v= t \<and> vars_term s \<inter> vars_term t = {}"

fun swaps :: "('v \<times> 'v) list \<Rightarrow> 'v \<Rightarrow> 'v" where
  "swaps [] = (\<lambda>x. x)" |
  "swaps ((x, y) # xys) = (swaps xys)(x := y, y := x)"

lemma swaps_Cons:
  assumes "swaps xys x = x"
    and "swaps xys y = y"
    and "\<forall>z. swaps xys z = x \<longrightarrow> z = x"
    and "\<forall>z. swaps xys z = y \<longrightarrow> z = y"
  shows "(Var \<circ> swaps ((x, y) # xys)) = (Var \<circ> swaps xys) \<circ>s swap x y"
proof (rule ext)
  fix z
  show "(Var \<circ> swaps ((x, y) # xys)) z = ((Var \<circ> swaps xys) \<circ>s swap x y) z"
    apply (auto simp add: assms(1, 2) subst_compose_def)
    apply (cases "swaps xys z = x")
    apply (insert assms(3), simp)
    apply (cases "swaps xys z = y")
    apply (insert assms(4), simp)
    unfolding swap_simps(3) [of "swaps xys z" x y]
    by simp
qed

lemma swaps_zip:
  assumes "x \<notin> set xs" and "x \<notin> set ys"
  shows "\<forall>z. swaps (zip xs ys) z = x \<longrightarrow> x = z"
  using assms by (induct xs ys rule: zip_induct) auto

lemma vren_Var_swaps:
  assumes "distinct xs" and "distinct ys" and "set xs \<inter> set ys = {}"
  shows "vren (Var \<circ> swaps (zip xs ys) :: ('f, 'v) subst)"
using assms
proof (induction xs ys rule: zip_induct)
  case (Cons_Cons x xs y ys)
  then have "distinct xs" and "x \<notin> set xs"
    and "distinct ys" and "y \<notin> set ys"
    and "set xs \<inter> set ys = {}"
    and "y \<notin> set xs"
    and "x \<notin> set ys" by auto
  with Cons_Cons.IH
    have "vren (Var \<circ> swaps (zip xs ys) :: ('f, 'v) subst)" by metis
  moreover have "swaps (zip xs ys) x = x"
    using `x \<notin> set xs` and `x \<notin> set ys`
    by (induct xs ys rule: zip_induct) auto
  moreover have "swaps (zip xs ys) y = y"
    using `y \<notin> set xs` and `y \<notin> set ys`
    by (induct xs ys rule: zip_induct) auto
  moreover have "\<forall>z. swaps (zip xs ys) z = x \<longrightarrow> z = x"
    using swaps_zip [OF `x \<notin> set xs` `x \<notin> set ys`] by force
  moreover have "\<forall>z. swaps (zip xs ys) z = y \<longrightarrow> z = y"
    using swaps_zip [OF `y \<notin> set xs` `y \<notin> set ys`] by force
  moreover have "vren (swap x y)" by (rule vren_subst)
  ultimately show ?case
    using vren_compose [OF `vren (Var \<circ> swaps _)` `vren (swap x y)`]
    by (simp only: zip.simps list.cases swaps_Cons)
qed (auto simp: o_def)

lemma swaps_vars:
  assumes "length xs \<le> length ys" and "x \<in> set xs"
    and "set xs \<inter> set ys = {}"
  shows "swaps (zip xs ys) x \<in> set ys"
  using assms by (induct xs ys arbitrary: x rule: zip_induct) auto

text {*If we have an infinite set of variables, we can always obtain a variable-disjoint
variant of a term.*}
lemma disj_variant_ex:
  fixes s :: "('f, 'v) term"
  assumes "infinite V"
  shows "\<exists>t. vars_term t \<subseteq> V \<and> s <v> t"
proof -
  from finite_distinct_list [OF vars_term_finite, of s] obtain xs
    where "set xs = vars_term s" and "distinct xs" by auto
  have "finite (vars_term s)" by (rule vars_term_finite)
  from disj_fresh_vars [OF `infinite V` this] obtain ys
    where "length ys = length xs" and "distinct ys"
    and "set ys \<subseteq> V - vars_term s" by blast
  then have "set ys \<subseteq> V" and "length xs \<le> length ys"
    and "set xs \<inter> set ys = {}"
    using `set xs = vars_term s` by auto
  def \<sigma> \<equiv> "(Var :: 'v \<Rightarrow> ('f, 'v) term) \<circ> swaps (zip xs ys)"
  def t \<equiv> "s \<cdot> \<sigma>"
  have "vars_term t \<subseteq> set ys"
    using swaps_vars [OF `length xs \<le> length ys` _ `set xs \<inter> set ys = {}`]
    unfolding t_def and \<sigma>_def and `set xs = vars_term s`
    by (induct s) (auto)
  with `set ys \<subseteq> V` have "vars_term t \<subseteq> V" by blast
  moreover have "s <v> t"
  proof (unfold disj_variant_def, rule)
    show "s =v= t"
    proof (rule variant_sym, unfold variant_def, rule+)
      show "vren \<sigma>"
        unfolding \<sigma>_def
        using vren_Var_swaps [OF `distinct xs` `distinct ys` `set xs \<inter> set ys = {}`] .
      show "t = s \<cdot> \<sigma>" by (simp add: t_def)
    qed
    show "vars_term s \<inter> vars_term t = {}"
      using `vars_term t \<subseteq> set ys`
      and `set xs = vars_term s`
      and `set xs \<inter> set ys = {}` by blast
  qed
  ultimately show ?thesis by blast
qed

end

